@@ -1,4 +1,4 @@
-{-# LANGUAGE BangPatterns, GADTs, FlexibleContexts, ScopedTypeVariables #-}
+{-# LANGUAGE BangPatterns, CPP, GADTs, FlexibleContexts, ScopedTypeVariables #-}
 -- |
 -- Module    : System.Random.MWC.Distributions
 -- Copyright : (c) 2012 Bryan O'Sullivan
@@ -41,7 +41,10 @@ import Control.Monad (liftM)
 import Control.Monad.Primitive (PrimMonad, PrimState)
 import Data.Bits ((.&.))
 import Data.Foldable (foldl')
-import Data.Traversable (Traversable,mapM)
+#if !MIN_VERSION_base(4,8,0)
+import Data.Traversable (Traversable)
+#endif
+import Data.Traversable (mapM)
 import Data.Word (Word32)
 import System.Random.MWC (Gen, uniform, uniformR)
 import qualified Data.Vector.Unboxed         as I
